#!/usr/bin/env python
"""
Get files using file metadata query

Examples::

  $ cta_prod_get_file_by_query MCCampaign=PROD5b particle=gamma-diffuse thetaP=20.0 phiP=180.0 site=LaPalma
"""
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.registerSwitch("", "Path=", "    Path to search for")
    Script.registerSwitch(
        "", "SE=", "    (comma-separated list of) SEs/SE-groups to be searched"
    )
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        [
            "metaspec: metadata index specification (of the form: "
            '"meta=value" or "meta<value", "meta!=value", etc.)'
        ],
        mandatory=False,
    )
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    import DIRAC
    from DIRAC import gLogger
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    from DIRAC.DataManagementSystem.Client.MetaQuery import (
        MetaQuery,
        FILE_STANDARD_METAKEYS,
    )
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup
    from DIRAC.Interfaces.API.Dirac import Dirac

    path = "/"
    seList = None
    for opt, val in Script.getUnprocessedSwitches():
        if opt == "Path":
            path = val
        elif opt == "SE":
            seList = resolveSEGroup(val.split(","))

    if seList:
        args.append(f"SE={','.join(seList)}")
    fc = FileCatalog()
    result = fc.getMetadataFields()
    if not result["OK"]:
        gLogger.error("Can not access File Catalog:", result["Message"])
        DIRAC.exit(-1)
    typeDict = result["Value"]["FileMetaFields"]
    typeDict.update(result["Value"]["DirectoryMetaFields"])
    # Special meta tags
    typeDict.update(FILE_STANDARD_METAKEYS)

    if len(args) < 1:
        print(f"Error: No argument provided\n{Script.scriptName}:")
        gLogger.notice(f"MetaDataDictionary: \n{str(typeDict)}")
        Script.showHelp(exitCode=1)

    mq = MetaQuery(typeDict=typeDict)
    result = mq.setMetaQuery(args)
    if not result["OK"]:
        gLogger.error("Illegal metaQuery:", result["Message"])
        DIRAC.exit(-1)
    metaDict = result["Value"]
    path = metaDict.pop("Path", path)

    result = fc.findFilesByMetadata(metaDict, path)
    if not result["OK"]:
        gLogger.error("Can not access File Catalog:", result["Message"])
        DIRAC.exit(-1)
    lfnList = sorted(result["Value"])

    gLogger.notice("Found lfns:")
    gLogger.notice("\n".join(lfn for lfn in lfnList))

    dirac = Dirac()
    for lfn in lfnList:
        gLogger.notice("Start downloading file", lfn)
        res = dirac.getFile(lfn)
        if not res["OK"]:
            gLogger.error("Error downloading file", lfn)
            return res["Message"]
        gLogger.notice("Successfully downloaded file", lfn)


if __name__ == "__main__":
    main()



"""
    Class to handle a workflow step described in CWL and composed of a commandLineTool
"""

__RCSID__ = "$Id$"

import schema_salad
from ruamel.yaml import YAML

yaml = YAML(typ="safe", pure=True)
import copy
from cwltool.load_tool import load_tool
from cwltool.context import LoadingContext, RuntimeContext
import cwltool.process
import cwltool.main
from CTADIRAC.ProductionSystem.CWL.CWLUtils import (
    check_if_file_input,
    add_fake_file,
    maybe_quote,
)


class WorkflowStep:
    """Composite class for workflow step (production step + job)"""

    def __init__(self):
        """Constructor"""
        self.command_line = ""
        self.process = None
        self.inputs = None
        self.sources = []
        self.sources_outputs = []

    def replace_step_input_by_source_output(self, step_input, source_output):
        if (source_output["type"] == "File") or (source_output["type"] == "Directory"):
            inp_class = source_output["type"]
            inp_location = source_output["outputBinding"]["glob"]
            self.inputs[step_input] = yaml.comments.CommentedMap(
                {"class": inp_class, "location": inp_location}
            )

    def add_source_output_to_tool_input(self, source_output):
        if (source_output["type"] == "File") or (source_output["type"] == "Directory"):
            inp_class = source_output["type"]
            inp_id = cwltool.process.shortname(source_output["_tool_entry"]["id"])
            inp_location = source_output["outputBinding"]["glob"]
            self.inputs[inp_id] = yaml.comments.CommentedMap(
                {"class": inp_class, "location": inp_location}
            )

    def load(self, input_cwl, input_yaml, sources=None, sources_outputs=None):
        loader = schema_salad.ref_resolver.Loader(
            {"@base": input_yaml, "path": {"@type": "@id"}}
        )
        self.inputs, _ = loader.resolve_ref(input_yaml)

        if sources:
            # Replace input by the output of the source
            for inp in sources:
                step_input = cwltool.process.shortname(inp["id"])
                for out in sources_outputs:
                    if out["id"] == inp["source"]:
                        self.replace_step_input_by_source_output(step_input, out)

        loading_context = LoadingContext({"strict": False, "debug": True})
        self.process = load_tool(input_cwl, loading_context)

    def verify_input(self):
        for inp in self.process.tool["inputs"]:
            short_input_name = cwltool.process.shortname(inp["id"])
            if short_input_name in self.inputs:
                continue
            elif short_input_name not in self.inputs and "default" in inp:
                self.inputs[short_input_name] = copy.copy(inp["default"])
            elif short_input_name not in self.inputs and inp["type"][0] == "null":
                continue
            else:
                raise ValueError(f"Missing inputs `{short_input_name}`")

    def build_job_command_line(self, runtime_context):
        for job in self.process.job(self.inputs, None, runtime_context):
            self.update_command_line(job)

    def update_command_line(self, job):
        for i in range(len(job.builder.bindings)):
            if isinstance(
                job.builder.bindings[i]["datum"], dict
            ):  # replace file path by its location
                if "location" in job.builder.bindings[i]["datum"]:
                    path = job.builder.bindings[i]["datum"]["path"]
                    location = job.builder.bindings[i]["datum"]["location"]
                    self.replace_path_in_command_line(job, path, location)
        self.command_line = " ".join(maybe_quote(arg) for arg in job.command_line)

    def replace_path_in_command_line(self, job, path, location):
        for k in range(len(job.command_line)):
            if path in job.command_line[k]:
                job.command_line[k] = job.command_line[k].replace(path, location)

    def get_command_line(self, sources=None, sources_outputs=None):
        """Get command line from a CommandLineTool description"""

        for entry in self.inputs.values():
            if check_if_file_input(entry):
                add_fake_file(entry["location"])

        self.verify_input()

        runtime_context = RuntimeContext()

        if sources:
            for out in sources_outputs:
                self.add_source_output_to_tool_input(out)

        self.build_job_command_line(runtime_context)
        #    for entry in self.inputs.values():
        #        if check_if_file_input(entry):
        #            remove_fake_file(entry["location"])
        return self



from CTADIRAC.Interfaces.API.CTAJob import CTAJob, MetadataDict


class CtapipeMergeJob(CTAJob):
    """Job extension class for ctapipe merging"""

    def __init__(self) -> None:
        super().__init__(we_type="merging")
        self.setType("Merging")
        self.setName("ctapipe_merge")
        self.prog_name = "ctapipe-merge"
        self.merged = 0
        self.data_level = 2
        self.output_extension = "merged.DL2.h5"

    def set_output_metadata(self, metadata: MetadataDict = {}) -> None:
        super().set_output_metadata(metadata)
        """For each stage of merging we add +1"""
        try:
            merged = metadata["merged"]["="]
        except (KeyError, TypeError):
            merged = metadata["merged"]
        self.output_metadata["merged"] = merged + 1

    def run_merge(self):
        step = self.setExecutable(
            "./dirac_ctapipe-merge_wrapper",
            arguments=f"--out_ext {self.output_extension} {self.options}",
            logFile="ctapipe_merge_Log.txt",
            modulesList=["cta_script"],
        )
        step["Value"]["name"] = "Step_ctapipe_merge"
        step["Value"]["descr_short"] = "Run ctapipe merge"

    def run_dedicated_software(self) -> None:
        self.run_merge()



from DIRAC.TransformationSystem.DB.TransformationDB import (
    TransformationDB as DIRACTransformationDB,
)


class TransformationDB(DIRACTransformationDB):
    """Add special features to TransformationDB"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def get_tasks_by_file_status(self, trans_id, file_status, task_status=None):
        query = (
            "select TaskID from TransformationTasks "
            f"where TransformationID={trans_id} and TaskID in ("
            "select TaskID from TransformationFiles "
            f"where TransformationID={trans_id} and Status='{file_status}' "
            f"group by TaskID)"
        )
        if task_status:
            formatted_statuses = ", ".join(f"'{status}'" for status in task_status)
            query += f" and ExternalStatus in ({formatted_statuses})"
        result_query = self._query(query + ";")
        return result_query



#!/usr/bin/env python

"""
  Get tasks attached to a transformation for a given TaskStatus
"""

__RCSID__ = "$Id$"

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


@Script()
def main():
    Script.registerSwitch("", "TransID=", "transformation ID")
    Script.registerSwitch("", "Status=", "task status")
    Script.parseCommandLine()

    transID = None
    status = None

    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "TransID":
            try:
                transID = int(switch[1])
            except Exception:
                gLogger.fatal("Invalid transID", switch[1])
        elif switch[0] == "Status":
            status = switch[1].capitalize()

    if not transID:
        Script.showHelp(exitCode=2)

    tc = TransformationClient()

    condDict = {"TransformationID": transID}
    if status:
        condDict.update({"ExternalStatus": status})

    res = tc.getTransformationTasks(condDict)

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(2)

    if len(res["Value"]) == 0:
        DIRAC.gLogger.notice(
            f"No tasks selected for transformation {transID} with status {status}"
        )

    for task in res["Value"]:
        DIRAC.gLogger.notice(task["TaskID"])


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Create datasets descriptions (in json format) for DL2 processing with ctapipe
The produced json files can be used as inputs to cta-prod-add-dataset

Usage example:
   cta-prod-create-dataset-description --MCCampaign=PROD5b --site=LaPalma --array_layout=Alpha --az=180 --zen=20.0 --ctapipe_ver=v0.19.0 --nsb=1
"""

__RCSID__ = "$Id$"

import json

import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.registerSwitch("", "MCCampaign=", "MCCampaign")
    Script.registerSwitch("", "site=", "site")
    Script.registerSwitch("", "array_layout=", "array_layout")
    Script.registerSwitch("", "az=", "azimuth angle")
    Script.registerSwitch("", "zen=", "zenith angle")
    Script.registerSwitch("", "div_ang=", "divergent angle")
    Script.registerSwitch("", "nsb=", "nsb")
    Script.registerSwitch("", "max_merged=", "max_merged")
    Script.registerSwitch("", "ctapipe_ver=", "ctapipe_version")
    switches, argss = Script.parseCommandLine(ignoreErrors=True)

    if not switches:
        Script.showHelp()

    # defaults
    MCCampaign = "PROD5b"
    site = "LaPalma"
    array_layout = "Alpha"
    phiP = 180.0
    thetaP = 20
    div_ang = None
    nsb = 1
    max_merged = 3
    analysis_prog_version = "v0.19.0"

    for switch in switches:
        if switch[0] == "MCCampaign":
            MCCampaign = switch[1]
        elif switch[0] == "site":
            site = switch[1]
        elif switch[0] == "array_layout":
            array_layout = switch[1]
        elif switch[0] == "az":
            if switch[1] == "0":
                phiP = 180.0
            elif switch[1] == "180":
                phiP = 0.0
            else:
                DIRAC.gLogger.error("az allowed values are: 0, 180")
        elif switch[0] == "zen":
            thetaP = float(switch[1])
        elif switch[0] == "div_ang":
            div_ang = switch[1]
        elif switch[0] == "nsb":
            nsb = int(switch[1])
        elif switch[0] == "max_merged":
            max_merged = int(switch[1])
        elif switch[0] == "ctapipe_ver":
            analysis_prog_version = switch[1]

    file_ext_list = [
        ".DL1ImgParDL2Geo.json",
        ".DL1ParDL2Geo.json",
        ".DL2GeoEneGam.json",
    ]
    file_name_list = []

    for file_ext in file_ext_list:
        if file_ext == ".DL1ImgParDL2Geo.json":
            analysis_prog = "ctapipe-merge"
            merged = 1
        if file_ext == ".DL1ParDL2Geo.json":
            analysis_prog = "ctapipe-merge"
            merged = max_merged
        if file_ext == ".DL2GeoEneGam.json":
            analysis_prog = "ctapipe-apply-models"
            merged = 0

        MDdict_common = {
            "analysis_prog": analysis_prog,
            "data_level": {"=": 2},
            "merged": {"=": merged},
            "sct": "False",
            "outputType": "Data",
        }

        for particle in ["gamma-diffuse", "proton", "gamma", "electron"]:
            MDdict = {
                "MCCampaign": MCCampaign,
                "site": site,
                "array_layout": array_layout,
                "particle": particle,
                "thetaP": {"=": int(thetaP)},
                "phiP": {"=": phiP},
                "nsb": {"=": nsb},
            }

            if div_ang:
                MDdict.update({"div_ang": {"=": div_ang}})

            MDdict.update({"analysis_prog_version": analysis_prog_version})

            if particle == "gamma-diffuse":
                if file_ext == ".DL2GeoEneGam.json":
                    split_list = ["test"]
                else:
                    split_list = ["train_en", "train_cl", "test"]
            elif particle == "proton":
                if file_ext == ".DL2GeoEneGam.json":
                    split_list = ["test"]
                else:
                    split_list = ["train_cl", "test"]
            elif particle in ["gamma", "electron"]:
                split_list = ["test"]

            for split in split_list:
                fields = []
                MDdict["split"] = split

                for key, value in MDdict.items():
                    if key == "thetaP":
                        fields.append("zen0" + str(value["="]))
                    elif key == "phiP":
                        if value["="] == 180:
                            az = "000"
                        elif value["="] == 0:
                            az = "180"
                        fields.append("az" + az)
                    elif key == "split":
                        fields.append(str(value) + "_merged")
                    elif key == "nsb":
                        fields.append("nsb0" + str(value["="]) + "x")
                    elif key == "div_ang":
                        fields.append("div" + str(value["="]))
                    elif key == "site":
                        fields.append(value.lower())
                    elif key in [
                        "MCCampaign",
                        "array_layout",
                        "particle",
                        "site",
                        "analysis_prog_version",
                    ]:
                        fields.append(value.lower())

                if particle in ["gamma", "electron"]:
                    MDdict.pop("split")

                MDdict.update(MDdict_common)
                json_string = json.dumps(MDdict)

                if div_ang:
                    file_name = (
                        "_".join(fields[0:4])
                        + "_"
                        + "".join(fields[4:8])
                        + "_"
                        + "_".join(fields[8:10])
                    )
                else:
                    file_name = (
                        "_".join(fields[0:4])
                        + "_"
                        + "".join(fields[4:7])
                        + "_"
                        + fields[7]
                        + "_"
                        + "_".join(fields[8:10])
                    )

                file_list = "_".join(fields[0:3]) + "_dataset.list"
                file_name_list.append(file_name.capitalize() + file_ext)
                # Write a json file with dataset query
                f = open(file_name.capitalize() + file_ext, "w")
                f.write(json_string.replace('"', "'") + "\n")
                f.close()

    # Add another dataset for all final DL2 files
    MDdict.pop("particle")
    MDdict["split"] = "test"
    json_string = json.dumps(MDdict)
    if div_ang:
        file_name = (
            "_".join(fields[0:3])
            + "_"
            + "".join(fields[4:8])
            + "_"
            + "_".join(fields[8:10])
        )
    else:
        file_name = (
            "_".join(fields[0:3])
            + "_"
            + "".join(fields[4:7])
            + "_"
            + fields[7]
            + "_"
            + "_".join(fields[8:10])
        )

    file_name_list.append(file_name.capitalize() + file_ext)
    f = open(file_name.capitalize() + file_ext, "w")
    f.write(json_string.replace('"', "'") + "\n")
    f.close()

    # Add dataset names to _dataset.list
    f = open(file_list.capitalize(), "w")
    for file_name in file_name_list:
        f.write(file_name + "\n")
    f.close()

    DIRAC.gLogger.notice(
        "list of dataset descriptions dumped into:", file_list.capitalize()
    )


####################################################
if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
  Check if files exist in the Catalog
  and dump the list of Found and Not Found files

  Usage:
  cta-dms-check-lfn [options] <ascii file with lfn list>
"""

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Base.Script import Script
from DIRAC import gLogger
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    if len(args) > 0:
        infile = args[0]
    else:
        Script.showHelp()

    lfns = read_inputs_from_file(infile)
    lfnsFound = []
    lfnsNotFound = []

    fc = FileCatalogClient()

    for lfn in lfns:
        res = fc.exists(lfn)
        if not res["OK"]:
            gLogger.error(res["Message"])

        if res["Value"]["Successful"][lfn]:
            gLogger.debug(f"{lfn} found in Catalog")
            lfnsFound.append(lfn)
        else:
            gLogger.debug(f"{lfn} not found in Catalog")
            lfnsNotFound.append(lfn)

    fname = infile + "_Found.list"
    f = open(fname, "w")
    for lfn in lfnsFound:
        f.write(lfn + "\n")
    f.close()
    gLogger.notice(f"{len(lfnsFound)} files have been dumped in {fname}")

    fname = infile + "_NotFound.list"
    f = open(fname, "w")
    for lfn in lfnsNotFound:
        f.write(lfn + "\n")
    f.close()
    gLogger.notice(f"{len(lfnsNotFound)} files have been dumped in {fname}")
    DIRAC.exit()


####################################################
if __name__ == "__main__":
    main()



import concurrent.futures
from collections import defaultdict
from typing import Any, Union
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.ReturnValues import DOKReturnType, DErrorReturnType
from DIRAC.Core.Base.AgentModule import AgentModule

from DIRAC.TransformationSystem.Client import (
    TransformationStatus,
    TransformationFilesStatus,
)
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import (
    JobMonitoringClient,
)
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from CTADIRAC.TransformationSystem.Utils.FailoverUtilities import FailoverUtilities
from CTADIRAC.Core.Utilities.return_values import s_report, DReportReturnType

AGENT_NAME = "Transformation/TransformationFailoverAgent"


class TransformationFailoverAgent(AgentModule, FailoverUtilities):
    """TransformationFailoverAgent:
    An Agent to manage failovers during Transformations
    Sends report, reassign files, complete or flush transfomations depending on configuration.
    Need a specific configuration for the agent:
    - TransformationStatus: Check the transformations with specific status (default: 'Active')
    - TransformationTypes: Check the transformations with specific types (default: None)
    - Report: Transformation type to report (default == TransformationTypes)
    - Complete: Complete transformation with specified type if conditions (default: None)
    - MaxFlush: Maximum times a transformation can be flushed (default: 2)
    - Flush: Flush transformation with specified type if conditions are fullfilled (default: None)
    - Reassign: Reassigned files to create new tasks when there is failed tasks in the transformation with specified type (default: None)
    - MaxReassign: Maximum number of reassigning files (default: 2)
    - maxThreadsInPool: Number of threads used by the agent
    - MailTo: Send report to sepcified mail adress
    - htmlReport: Return the report as html format (default: True)
    """

    def __init__(self, *args, **kwargs) -> None:
        AgentModule.__init__(self, *args, **kwargs)
        FailoverUtilities.__init__(self, agent_name=AGENT_NAME)

        self.trigger_job_status: list[str] = [JobStatus.FAILED, JobStatus.KILLED]
        self.failed_final_task_status: list[str] = [JobStatus.FAILED, JobStatus.KILLED]
        self.jobs_final_status: list[str] = [
            JobStatus.DONE,
            JobStatus.COMPLETED,
            JobStatus.FAILED,
            JobStatus.KILLED,
        ]
        self.default_transformation_status: list[str] = [TransformationStatus.ACTIVE]
        self.file_status_to_check = TransformationFilesStatus.ASSIGNED
        self.trans_with_no_input: list = ["MCSimulation"]
        self.html = True

        self.job_mon = JobMonitoringClient()

        self.get_cs_options()

        self.mail_to = ""
        self.mail_from = "noreply@ctao.dirac"
        self.subject = "[TransformationFailoverAgent]"

        self.reassign_count = {}
        self.flush_count = {}

    def initialize(self) -> DOKReturnType[None]:
        """Agent's initialisation"""
        self.mail_from = self.am_getOption("MailFrom", self.mail_from)
        self.mail_to = self.am_getOption("MailTo", self.mail_to)

        max_number_of_threads = self.am_getOption("maxThreadsInPool", 15)
        gLogger.info(f"Multithreaded with {max_number_of_threads} threads")
        self.thread_pool_executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=max_number_of_threads
        )
        return S_OK()

    def get_cs_options(self) -> None:
        """Re initialize before execution"""

        self.html = self.am_getOption("htmlReport", self.html)
        self.transformation_status = self.am_getOption(
            "TransformationStatus", self.default_transformation_status
        )

        self.transformation_types = self.am_getOption("TransformationTypes", [])
        self.transformation_to_report = self.am_getOption(
            "Report", self.transformation_types
        )
        self.transformations_to_complete = self.am_getOption("Complete", [])
        self.transformation_to_flush = self.am_getOption("Flush", [])
        self.transformation_to_reassign = self.am_getOption("Reassign", [])

        self.max_flush = self.am_getOption("MaxFlush", 2)
        self.max_reassign = self.am_getOption("MaxReassign", 2)

    def should_complete_transformation(
        self, trans_type: str, trans_id: int, parent_id: int
    ) -> bool:
        """Check if transformation should be completed.
        i.e. has input, is processed and has no parent
        or parent is processed"""
        not_completed_message: str = (
            f"[{trans_id}] Do not complete transformation: requirement not satisfied"
        )
        if (
            trans_type in self.trans_with_no_input
            or trans_type not in self.transformations_to_complete
            or parent_id is None
        ):
            gLogger.info(not_completed_message)
            return False

        if self.is_transformation_processed(trans_id) and (
            parent_id == -1
            or (parent_id != -1 and self.is_transformation_processed(parent_id))
            or (parent_id != -1 and self.get_parent_type(parent_id) == "MCSimulation")
        ):
            gLogger.info(f"[{trans_id}] Completing transformation...")
            return True

        gLogger.info(not_completed_message)
        return False

    def execute_transformation_check(
        self, trans_id: int, trans_info_dict: dict
    ) -> Union[DReportReturnType, DErrorReturnType, DOKReturnType[dict], Any]:
        """Check transformation before creating a report
        if necessary flush and reassign files"""
        gLogger.info(f"[{trans_id}] Treating Transformation...")

        trans_type = trans_info_dict["Type"]
        parent_id = self.get_parent_transformation_id(trans_id)
        reassigned_lfns: dict = {}
        no_report: str = f"[{trans_id}] Report not created"

        tasks_infos = self.get_transformation_tasks(trans_id)["Value"]

        jobs_id: list = [
            job["ExternalID"] for job in tasks_infos if job["ExternalID"] != "0"
        ]
        res_jobs_attributes = self.job_mon.getJobsParameters(
            jobs_id, self.attributes_of_interest
        )
        if not res_jobs_attributes["OK"]:
            return s_report(False, f"{res_jobs_attributes}")

        jobs_attributes: dict = res_jobs_attributes["Value"]

        is_there_updated_tasks: bool = self.update_tasks_status(
            trans_id, jobs_attributes, tasks_infos
        )
        if is_there_updated_tasks:
            tasks_infos = self.get_transformation_tasks(trans_id)["Value"]

        # Do nothing if all files are processed and complete transformation:
        if self.should_complete_transformation(trans_type, trans_id, parent_id):
            return self.complete_transformation(trans_id)

        if (
            trans_type not in self.trans_with_no_input
            and self.is_transformation_processed(trans_id)
        ):
            return s_report(False, f"[{trans_id}] is processed")

        # Check if transformation needs to be flushed
        has_been_flushed = False
        if (
            trans_type in self.transformation_to_flush
            and self.flush_count.setdefault(trans_id, 0) < self.max_flush
        ):
            group_size: int = trans_info_dict["GroupSize"]
            has_been_flushed: bool = self.flush_transformation(
                trans_id, group_size, parent_id, tasks_infos
            )

        # Set files status to unused if associated tasks failed or killed
        if trans_type in self.transformation_to_reassign and not has_been_flushed:
            reassigned_lfns = self.reassign_lfn(trans_id)

        if trans_type not in self.transformation_to_report:
            return s_report(
                False, f"{no_report}: {trans_type} type should not be reported"
            )

        # Create Report on transformation
        report_result = self.parse_transformation_jobs_attributes(
            trans_id, jobs_attributes
        )

        if not report_result["OK"]:
            return report_result

        if reassigned_lfns:
            report_lfns: str = self.create_reassigned_lfn_report(
                trans_id, reassigned_lfns
            )
            return s_report(True, report_result["Value"] + report_lfns)

        return s_report(True, report_result["Value"])

    def update_tasks_status(
        self, trans_id: int, jobs_attributes: dict, tasks_infos: list
    ) -> bool:
        """Update task status in TransformationDB"""
        tasks_status_dict = {}
        has_been_updated = False
        for task in tasks_infos:
            tasks_status_dict[int(task["ExternalID"])] = {
                "ExternalStatus": task["ExternalStatus"],
                "TransformationID": task["TransformationID"],
                "TaskID": task["TaskID"],
            }
        for job_id, job_info in jobs_attributes.items():
            try:
                if tasks_status_dict[job_id]["ExternalStatus"] != job_info["Status"]:
                    gLogger.verbose(
                        f"[{trans_id}] Setting new task [{tasks_status_dict[job_id]['TaskID']}] status"
                    )
                    res = self.set_task_status(
                        tasks_status_dict[job_id]["TransformationID"],
                        tasks_status_dict[job_id]["TaskID"],
                        job_info["Status"],
                    )
                    if not res["OK"]:
                        gLogger.error(f"[{trans_id}] {res}")
                        continue
                    else:
                        gLogger.verbose(
                            f"[{trans_id}] Task [{tasks_status_dict[job_id]['TaskID']}] status changed: {tasks_status_dict[job_id]['ExternalStatus']} => {job_info['Status']}"
                        )
                        has_been_updated = True
            except KeyError as ke:
                gLogger.error(f"[{trans_id}] {ke}")
                continue
        return has_been_updated

    def get_lfns_to_assign(self, trans_id: int, failed_tasks: list) -> dict[int, str]:
        """Get LFNs to assign based on failed tasks"""
        lfns_to_assign = {}
        if not isinstance(failed_tasks, list):
            failed_tasks_list = [task[0] for task in failed_tasks]
        else:
            failed_tasks_list = failed_tasks
        res_trans_files = self.get_transformation_files(
            {"TransformationID": trans_id, "TaskID": failed_tasks_list}
        )
        if res_trans_files["OK"]:
            for trans_file in res_trans_files["Value"]:
                lfn = trans_file["LFN"]
                if (
                    lfn not in self.reassign_count
                    or self.reassign_count[lfn] <= self.max_reassign
                ):
                    index = trans_file["TaskID"]
                    lfns_to_assign.setdefault(index, [])
                    lfns_to_assign[index] += [lfn]
                    self.update_reassign_count(lfn)
        else:
            gLogger.error(
                f"[{trans_id}] Failed to getTransformationFiles: {res_trans_files}"
            )
        return lfns_to_assign

    def update_reassign_count(self, lfn: str) -> None:
        """Update reassign count for the specified LFN"""
        if lfn in self.reassign_count:
            self.reassign_count[lfn] += 1
        else:
            self.reassign_count[lfn] = 1

    def reassign_lfn(self, trans_id: int) -> dict[int, list]:
        """Set LFNs to Unused if failed or killed tasks"""
        lfns_to_assign: dict = {}
        reassigned_lfns: dict = {}
        res_assigned_tasks_status = self.trans_client.get_tasks_by_file_status(
            trans_id, self.file_status_to_check, self.failed_final_task_status
        )
        if res_assigned_tasks_status["OK"]:
            failed_tasks = res_assigned_tasks_status["Value"]
            if failed_tasks:
                lfns_to_assign = self.get_lfns_to_assign(trans_id, failed_tasks)
                lfns = [
                    lfn for tasks_lnfs in lfns_to_assign.values() for lfn in tasks_lnfs
                ]
                gLogger.verbose(
                    f"[{trans_id}] Changing files status to unused on lnfs: {lfns_to_assign}"
                )
                res = self.trans_client.setFileStatusForTransformation(
                    trans_id,
                    newLFNsStatus=TransformationFilesStatus.UNUSED,
                    lfns=lfns,
                    force=True,
                )
                if not res["OK"]:
                    gLogger.error(f"[{trans_id}] {res}")
                else:
                    gLogger.info(f"[{trans_id}] Files status changed successfully")
                    reassigned_lfns = lfns_to_assign
        else:
            gLogger.error(f"{trans_id}: {res_assigned_tasks_status}")

        return reassigned_lfns

    def flush_transformation(
        self, trans_id: int, group_size: dict, parent_id: int, tasks_infos: list
    ) -> bool:
        """Flush transformation if necessary conditions are fullfilled"""
        parent_status = None
        res_parent_trans = self.trans_client.getTransformation(parent_id)
        if res_parent_trans["OK"]:
            parent_status = res_parent_trans["Value"]["Status"]

        flush_trans, message = self.should_flush_transformation(
            trans_id, parent_id, parent_status, group_size, tasks_infos
        )
        if flush_trans:
            self.execute_flush(trans_id, message)
        return flush_trans

    def execute_flush(self, trans_id: int, message: str):
        """Actually flush the transformation"""
        res_flush = self.trans_client.setTransformationParameter(
            trans_id, paramName="Status", paramValue=TransformationStatus.FLUSH
        )
        if res_flush["OK"]:
            self.send_mail(
                subject=self.subject + f"Flush Transformation {trans_id}",
                message=message,
            )
            self.flush_count[trans_id] += 1
            gLogger.info(f"[{trans_id}] {message}")
        else:
            gLogger.error(res_flush)

    def should_flush_transformation(
        self,
        trans_id: int,
        parent_id: int,
        parent_status: str,
        group_size: int,
        tasks_infos: list,
    ) -> tuple[bool, str]:
        flush_trans = False
        count_unused = 0
        message = ""
        if parent_id != -1 and parent_status != TransformationStatus.COMPLETED:
            gLogger.info(
                f"[{trans_id}] Parent Transformation {parent_id} is not complete yet."
            )
        else:
            count_unused: int = self.get_count_size(trans_id)
            if count_unused != 0 and group_size > count_unused:
                flush_trans = True
                message: str = (
                    "Nb of files with Status 'Unused' "
                    f"({count_unused}) < GroupSize ({group_size})\n"
                    f"The Transformation {trans_id} has been Flushed"
                )
            elif count_unused != 0 and group_size <= count_unused and tasks_infos:
                task_in_final_state = True
                for task in tasks_infos:
                    if task["ExternalStatus"] not in self.jobs_final_status:
                        task_in_final_state = False
                        break
                flush_trans = task_in_final_state
                message: str = (
                    "Nb of files with Status 'Unused' "
                    f"({count_unused}) > GroupSize ({group_size})"
                    "and all jobs are in final states.\n"
                    f"The Transformation {trans_id} has been Flushed"
                )
        return flush_trans, message

    def get_count_size(self, trans_id: int) -> int:
        unused_trans_files = self.get_transformation_files(
            {"TransformationID": trans_id, "Status": "Unused"}
        )
        if not unused_trans_files["OK"]:
            gLogger.error(f"[{trans_id}] {unused_trans_files}")
            return 0
        return len(unused_trans_files["Value"])

    def parse_transformation_jobs_attributes(
        self, trans_id: int, jobs_attributes: list, create_report: bool = True
    ) -> Union[DOKReturnType[dict], Any, DErrorReturnType]:
        """Create a transformation report giving the number of jobs
        with a given application status by sites"""

        count_status_type: dict = {}
        count_minor_status: dict = {}
        count_status: dict = {}

        total_nb_jobs: int = len(jobs_attributes.keys())

        # Count jobs by status type:
        for job_id, job_info in jobs_attributes.items():
            job_attributes_tuple = tuple(
                job_info[attr] for attr in self.attributes_of_interest
            )

            status_counts = count_status_type.get(job_attributes_tuple, 0)
            count_status_type[job_attributes_tuple] = status_counts + 1

            count_minor_status[job_info["MinorStatus"]] = (
                count_minor_status.get(job_info["MinorStatus"], 0) + 1
            )
            count_status[job_info["Status"]] = (
                count_status.get(job_info["Status"], 0) + 1
            )
        if not count_status_type:
            return S_ERROR(f"[{trans_id}] Empty Transformation")

        if create_report:
            transformation_report: str = self.create_transformation_jobs_report(
                trans_id,
                count_status_type,
                count_minor_status,
                count_status,
                total_nb_jobs,
            )
            return S_OK(transformation_report)
        else:
            return S_OK()

    def create_threads_for_each_transformations(self, transformations) -> dict:
        """Create the future threads to check each transformation"""
        future_to_trans_id: dict = {}
        for trans_id, trans_info_dict in transformations["Value"].items():
            trans_type = trans_info_dict["Type"]
            trans_id = int(trans_id)
            try:
                future = self.thread_pool_executor.submit(
                    self.execute_transformation_check, trans_id, trans_info_dict
                )
                future_to_trans_id[future] = (trans_id, trans_type)
            except Exception.error as e:
                gLogger.info(e)
        return future_to_trans_id

    def before_executing(self) -> None:
        self.get_cs_options()

    def execute(self) -> Union[DErrorReturnType, DOKReturnType[None]]:
        """Execution in one agent's cycle"""
        self.before_executing()

        transformations = self.get_eligible_transformation(
            self.transformation_status, self.transformation_types
        )
        if not transformations["OK"]:
            gLogger.error("Failure to get transformations")
            return S_ERROR(transformations["Message"])

        future_to_trans_id = self.create_threads_for_each_transformations(
            transformations
        )

        # Get results:
        # TODO: Simplify the output parsing
        transformation_reports: dict[str, dict[str, str]] = defaultdict(dict)
        for future in concurrent.futures.as_completed(future_to_trans_id):
            trans_id, trans_type = future_to_trans_id[future]
            try:
                result = future.result()
                if result["OK"]:
                    trans_report = result["Value"]
                    if trans_report["Report"]:
                        transformation_reports[trans_type][trans_id] = trans_report[
                            "Message"
                        ]
                        gLogger.info(f"Report for transformation {trans_id} created")
                    else:
                        gLogger.info(trans_report["Message"])
                else:
                    gLogger.error(result)
            except Exception as exc:
                gLogger.error(f"{trans_id} generated an exception: {exc}")
                gLogger.debug(future.result())
            else:
                gLogger.info(f"Processed transformation {trans_id}")

        if transformation_reports:
            message: str = self.create_mail_message(
                transformation_reports, html=self.html
            )
            self.send_mail(
                subject=(
                    f"{self.subject} Report on "
                    f"{self.transformation_status} transformations"
                ),
                message=message,
                html=self.html,
            )
        return S_OK()

    def finalize(self) -> DOKReturnType[None]:
        """Final step before shutdown the agent"""
        try:
            self.thread_pool_executor.shutdown()
        except Exception.error as e:
            gLogger.error(f"Threads shutdown failed: {e}")
        else:
            gLogger.info("Finishing threads")
        return S_OK()



#!/usr/bin/env python
"""
Create datasets descriptions (in json format) for R1->DL1 processing with EventDisplay
The produced json files can be used as inputs to cta-prod-add-dataset

Usage example:
   cta-prod-create-evndisp-dataset-description --MCCampaign=PROD6 --site=LaPalma --array_layout=Prod6-Hyperarray --az=180 --zen=20.0 --evndisp_ver=v5.13.2 --nsb=1
"""

__RCSID__ = "$Id$"

import json

import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.registerSwitch("", "MCCampaign=", "MCCampaign")
    Script.registerSwitch("", "site=", "site")
    Script.registerSwitch("", "array_layout=", "array_layout")
    Script.registerSwitch("", "az=", "azimuth angle")
    Script.registerSwitch("", "zen=", "zenith angle")
    Script.registerSwitch("", "nsb=", "nsb")
    Script.registerSwitch("", "evndisp_ver=", "evndisp version")
    switches, argss = Script.parseCommandLine(ignoreErrors=True)

    if not switches:
        Script.showHelp()

    # defaults
    MCCampaign = "PROD6"
    site = "LaPalma"
    array_layout = "Prod6-Hyperarray"
    phiP = 180.0
    thetaP = 20
    nsb = 1
    analysis_prog_version = "v5.13.2"

    for switch in switches:
        if switch[0] == "MCCampaign":
            MCCampaign = switch[1]
        elif switch[0] == "site":
            site = switch[1]
        elif switch[0] == "array_layout":
            array_layout = switch[1]
        elif switch[0] == "az":
            if switch[1] == "0":
                phiP = 180.0
            elif switch[1] == "180":
                phiP = 0.0
            else:
                DIRAC.gLogger.error("az allowed values are: 0, 180")
        elif switch[0] == "zen":
            thetaP = float(switch[1])
        elif switch[0] == "nsb":
            nsb = int(switch[1])
        elif switch[0] == "evndisp_ver":
            analysis_prog_version = switch[1]

    MDdict = {
        "MCCampaign": MCCampaign,
        "site": site,
        "array_layout": array_layout,
        "thetaP": {"=": int(thetaP)},
        "phiP": {"=": phiP},
        "nsb": {"=": nsb},
        "analysis_prog": "evndisp",
        "analysis_prog_version": analysis_prog_version,
        "data_level": {"=": 1},
        "sct": "False",
        "outputType": "Data",
    }

    file_name_list = []

    for particle in ["gamma-diffuse", "proton", "gamma", "electron"]:
        MDdict.update({"particle": particle})

        fields = []

        for key, value in MDdict.items():
            if key == "thetaP":
                fields.append("zen0" + str(value["="]))
            elif key == "phiP":
                if value["="] == 180:
                    az = "000"
                elif value["="] == 0:
                    az = "180"
                fields.append("az" + az)
            elif key == "nsb":
                fields.append("nsb0" + str(value["="]) + "x")
            elif key == "site":
                fields.append(value.lower())
            elif key in [
                "MCCampaign",
                "array_layout",
                "particle",
                "site",
                "analysis_prog_version",
            ]:
                fields.append(value.lower())

        json_string = json.dumps(MDdict)

        file_name = (
            "_".join(fields[0:3])
            + "_"
            + fields[7]
            + "_"
            + "".join(fields[3:6])
            + "_"
            + fields[6]
            + "_evndisp_DL1.json"
        )

        file_list = (
            "_".join(fields[0:3]) + "_" + "_".join(fields[3:6]) + "_dataset.list"
        )
        file_name_list.append(file_name.capitalize())
        # Write a json file with dataset query
        f = open(file_name.capitalize(), "w")
        f.write(json_string.replace('"', "'") + "\n")
        f.close()

    # Add dataset names to _dataset.list
    f = open(file_list.capitalize(), "w")
    for file_name in file_name_list:
        f.write(file_name + "\n")
    f.close()

    DIRAC.gLogger.notice(
        "list of dataset descriptions dumped into:", file_list.capitalize()
    )


####################################################
if __name__ == "__main__":
    main()



#!/usr/bin/env python

__RCSID__ = "$Id$"

# generic imports
import os

# DIRAC imports
from DIRAC.Core.Base.Script import Script
import DIRAC

Script.parseCommandLine(ignoreErrors=True)

from DIRAC import gLogger
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient


@Script()
def main():
    fcc = FileCatalogClient()
    dirac = Dirac()
    resJDL = dirac.getJobJDL(os.environ["JOBID"])

    idata = resJDL["Value"]["InputData"]
    if isinstance(idata, str):
        idata = []
        if "LFN" in resJDL["Value"]["InputData"]:
            idata.append(resJDL["Value"]["InputData"].split("LFN:")[1])
        else:
            idata.append(resJDL["Value"]["InputData"])
    else:
        idata = resJDL["Value"]["InputData"]

    for lfn in idata:
        gLogger.notice(f"Checking input Log file:\n {lfn} ")
        file_name = os.path.basename(lfn)

        # List the content of the tar file
        res = list(os.popen("tar -tf " + file_name + " '*log.gz'"))
        tardir = res[0].split("/")[0]
        # Extract the tar file
        code = os.system("tar -xvf " + file_name)
        if code != 0:
            gLogger.error(f"Unable to untar Log file:\n {file_name}")
            DIRAC.exit(-1)
        os.system("gunzip -r " + tardir)
        res = list(
            os.popen("grep camera_CTA-LST -r " + tardir + " | grep file | wc -l")
        )
        if res[0].strip() != "4":
            data_lfn = lfn.replace("Log", "Data").replace("log_hist.tar", "simtel.zst")
            gLogger.error(f"Buggy Data file:\n {data_lfn}")
            gLogger.notice("Setting nsb=-1")
            res = fcc.setMetadata(data_lfn, {"nsb": -1})
            if not res["OK"]:
                return res

        code = os.system("rm -Rf " + tardir)
        if code != 0:
            gLogger.error("Unable to remove tar directory")
            DIRAC.exit(-1)
        code = os.system("rm " + file_name)
        if code != 0:
            gLogger.error(f"Unable to remove Log file:\n {file_name}")
            DIRAC.exit(-1)

    DIRAC.exit()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
""" Data management script for production
    create DFC MetaData structure put and register files in DFC
    should work for corsika, simtel and EventDisplay output
"""

__RCSID__ = "$Id$"

# generic imports
import os
import glob
import json

# DIRAC imports
import DIRAC
from DIRAC.Core.Base.Script import Script

# CTADIRAC imports
from CTADIRAC.Core.Utilities.tool_box import run_number_from_filename
from CTADIRAC.Core.Workflow.Modules.ProdDataManager import ProdDataManager
from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file

Script.parseCommandLine()


@Script()
def main():
    """simple wrapper to put and register all production files

    Keyword arguments:
    args -- a list of arguments in order []
    """
    args = Script.getPositionalArgs()
    metadata = args[0]
    file_metadata = args[1]
    base_path = args[2]
    output_pattern = args[3]
    package = args[4]
    program_category = args[5]
    catalogs = args[6]
    output_type = args[7]
    if len(args) == 9:
        update_ts = True
    else:
        update_ts = False

    exit_status = 0

    # Load catalogs
    catalogs_json = json.loads(catalogs)

    # Init ProdDataManager
    prod_dm = ProdDataManager(catalogs_json)

    # If wms execution don't update TS file status
    if prod_dm.TransformationID == "0000":
        update_ts = False

    # Define list for update TS File Status
    out_lfn_to_reassign_list = []
    to_reassign_lfn_list = []
    in_lfn_list = prod_dm.InputData

    if not isinstance(in_lfn_list, list):
        in_lfn_list = in_lfn_list.split("LFN:")

    # Create MD structure
    result = prod_dm.createMDStructure(
        metadata, base_path, program_category, output_type
    )
    if result["OK"]:
        path = result["Value"]
    else:
        return result

    # Check the content of the output directory
    res = prod_dm._checkemptydir(output_pattern)
    if not res["OK"]:
        DIRAC.gLogger.warn(res["Message"])
        to_reassign_lfn_list = in_lfn_list.copy()

    # Dump the list of output LFNs
    file = open("output_lfns.txt", "w")

    # Loop over each file and upload and register
    for localfile in glob.glob(output_pattern):
        file_name = os.path.basename(localfile)
        # Check run number, assign one as file metadata if needed
        fmd_dict = json.loads(file_metadata)
        try:
            run_number = run_number_from_filename(file_name, package)
        except BaseException:
            run_number = -9999
            DIRAC.gLogger.notice("Could not get a correct run number, assigning -9999")
        fmd_dict["runNumber"] = "%08d" % int(run_number)
        # get the output file path
        run_path = prod_dm._getRunPath(fmd_dict)
        lfn = os.path.join(path, output_type, run_path, file_name)
        fmd_json = json.dumps(fmd_dict)
        result = prod_dm.putAndRegister(lfn, localfile, fmd_json, package)
        if not result["OK"]:
            out_lfn_to_reassign_list.append(lfn)
            exit_status = 1
        else:
            file.write(lfn)
            file.write("\n")
    file.close()

    ### Add here the update of TS Status
    if update_ts:
        map_lfn = {}
        map_to_reassign_lfn = {}
        out_lfn_list = read_inputs_from_file("output_lfns.txt")
        problematic_file_list = []
        if os.path.exists("problematic_files.txt"):
            problematic_file_list = read_inputs_from_file("problematic_files.txt")
        problematic_lfn_list = []

        for in_lfn in in_lfn_list:
            in_file_name = os.path.basename(in_lfn)
            if in_file_name in problematic_file_list:
                problematic_lfn_list.append(in_lfn)

        for out_lfn in out_lfn_list:
            out_file_name = os.path.basename(out_lfn)
            out_run_number = run_number_from_filename(out_file_name, package)
            out_run_number = "run%06d" % int(out_run_number)
            for in_lfn in in_lfn_list:
                in_file_name = os.path.basename(in_lfn)
                if out_run_number in in_file_name:
                    map_lfn[out_lfn] = in_lfn

        for out_lfn in out_lfn_to_reassign_list:
            out_file_name = os.path.basename(out_lfn)
            out_run_number = run_number_from_filename(out_file_name, package)
            out_run_number = "run%06d" % int(out_run_number)
            for in_lfn in in_lfn_list:
                in_file_name = os.path.basename(in_lfn)
                if out_run_number in in_file_name:
                    map_to_reassign_lfn[out_lfn] = in_lfn

        DIRAC.gLogger.notice("Mapping output/input")
        for key, value in map_lfn.items():
            DIRAC.gLogger.notice(
                f"{os.path.basename(key)} mapped to {os.path.basename(value)}"
            )
            DIRAC.gLogger.notice(f"Setting file status to PROCESSED {value}")
            res = prod_dm.setTransformationFileStatus(value, "PROCESSED")
            if not res["OK"]:
                DIRAC.gLogger.warn(f"Failed to set file status to PROCESSED {value}")

        for lfn in problematic_lfn_list:
            DIRAC.gLogger.notice(f"Setting file status to PROBLEMATIC {lfn}")
            res = prod_dm.setTransformationFileStatus(lfn, "PROBLEMATIC")
            if not res["OK"]:
                DIRAC.gLogger.warn(f"Failed to set file status to PROBLEMATIC {lfn}")

        for key, value in map_to_reassign_lfn.items():
            if value not in problematic_lfn_list:
                DIRAC.gLogger.notice(f"Setting file status to UNUSED {value}")
                res = prod_dm.setTransformationFileStatus(value, "UNUSED")
                if not res["OK"]:
                    DIRAC.gLogger.warn(f"Failed to set file status to UNUSED {value}")

        for lfn in to_reassign_lfn_list:
            DIRAC.gLogger.notice(f"Setting file status to UNUSED {lfn}")
            res = prod_dm.setTransformationFileStatus(value, "UNUSED")
            if not res["OK"]:
                DIRAC.gLogger.warn(f"Failed to set file status to UNUSED {lfn}")

    DIRAC.exit(exit_status)


####################################################
if __name__ == "__main__":
    main()
